<div class="dropdown">
    <button type="button" class="btn btn-{% if (build_dev_html|tobool) %}danger{% else %}primary{% endif %} btn-sm navbar-btn dropdown-toggle" id="dLabelMore" data-toggle="dropdown">
        v{{ release }}
        <span class="caret"></span>
    </button>
    <div class="dropdown-menu list-group-flush py-0" aria-labelledby="dLabelMore">
        {%- for ver, txt in versions_dropdown.items() %}
        <a class="list-group-item list-group-item-action py-1" href="https://mne.tools/mne-bids/{{ ver }}/index.html">{{ txt }}</a>
        {%- endfor %}
    </div>
</div>
